//八皇后问题的验证程序
// 要点：
//     1 归纳谓词描述皇后在棋盘上正确摆放的条件
//     2 函数 EightQueens 利用全局数组记录一个解的状态，并且找到一个，输出一个。
//       函数没有形参和返回值，所以函数协议的前后条件均为\true。
//     3 循环不变式给出了判断解的断言
//     4 程序代码参见（源码分享 - C语言中文网c.biancheng.net/cpp/u/yuanma/）

#include <stdbool.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>

#define QUEENS 8 //定义结果数组的大小，也就是皇后的数目

typedef int SRowCol[QUEENS + 1]; // 记录皇后所在的行列位置，从1开始算起。
SRowCol a={0};                       // a[i] = j 表示第i行的皇后放在第j列

/*@ 
    // 皇后不在相同的行列上
    inductive NotSameColumn(SRowCol a,integer first, integer n) =
        n == first // first==0
        || 
        n >= first+1  // 第n位皇后
        && NotSameColumn(a,first, n-1) // n-1位皇后不在同一列
        && (\forall integer i:[first+1..n-1].a[i]!=a[n])  // 皇后n与其他各位皇后也不再同一列上
        ;
    // 皇后不在相同的对角线上
    inductive NotSameDiagonal(SRowCol a,integer first, integer n) =
        n == first 
        || 
        n >= first+1 
        && NotSameDiagonal(a,first, n-1) 
        && (\forall integer i:[first+1..n-1].a[n]!=a[i]-(i-n) // 正斜线
            && a[n]!=a[i]+(i-n)                               // 反斜线
           )
        ;
*/

//@ ghost _Bool arithmetic_overflow;

/*@ requires \true;
    assigns a, arithmetic_overflow;
    ensures \true;
    exits \exit_status == 1 && arithmetic_overflow == true;
*/
/*
    采用回溯策略生成所有的可能解的基本思想是：
    1. 从a[1] = 1开始生成第一个解；
    2. 从后向前调整皇后的位置，求出下一个解，直至找到所有的解。
       所有情况都探测完毕的条件是：a[1]==QUEENS时flag==0，需要重置a[1]的值寻求下一个解。
*/
void EightQueens()
{ // 采用回溯策略生成所有的可能解，并打印输出
    int i, k;
    int flag;           // flag==1:当前行皇后摆放未出现冲突
    int count = 0;      // count:符合条件的八皇后摆放方法数目。
    int not_finish = 1; // not_finished==1:尚未完成搜索。
                        // not_finished==0的条件是：flag==0，需要重置a[i]，而这时i==1并且a[i]==QUEENS。

    i = 1;    //i表示行号，初始时摆放第1行的皇后
    a[1] = 1; //从第1列开始尝试
    printf("The possible configuration of 8 queens are:\n");
    /*@ loop invariant 
        1 <= i <= QUEENS && (\forall integer j:[1..i].0 < a[j] <= QUEENS) 
        && not_finish==1 && NotSameColumn(a,0,i-1) && NotSameDiagonal(a,0,i-1) 
        && INT_MAX >= count >= 0 
      || 
        not_finish == 0
        ;
    */
    while (not_finish == 1)
    { //not_finish=l:处理尚未结束
        /*@ loop invariant 
            1<=i<=QUEENS && (\forall integer j:[1..i].0 < a[j] <= QUEENS) 
            && not_finish==1 && NotSameColumn(a,0,i-1) && NotSameDiagonal(a,0,i-1) 
            && INT_MAX >= count >= 0 
          ||
            i==QUEENS+1 && (\forall integer j:[1..QUEENS].0 < a[j] <= QUEENS)  // 处理到第QUEENS个元素
            && not_finish==1 && NotSameColumn(a,0,QUEENS) && NotSameDiagonal(a,0,QUEENS) 
            && INT_MAX >= count >= 0 
          ||
            not_finish==0
            ;                                                     // 处理结束
        */
        while (not_finish == 1 && i <= QUEENS)
        { //处理尚未结束且还没处理到第QUEENS个元素
            flag = 1;
            /*@ loop invariant 
                1<=i<=QUEENS && 1<=k<=i && (\forall integer j:[1..i].0 < a[j] <= QUEENS) 
                && not_finish==1 && NotSameColumn(a,0,i-1) && NotSameDiagonal(a,0,i-1) 
                && INT_MAX >= count >= 0 
                && (flag==1 && (\forall integer j:[1..k-1].a[j]!=a[i]) || flag==0 && i > 1 && a[k-1]==a[i])
                ; 
            */
            for (k = 1; flag == 1 && k < i; k = k + 1)
            { // 判断是否有多个皇后在同一列上。
                if (a[k] == a[i])
                { // 注意：i==1时，k<i条件不成立，因此flag=0时，必有i>1。
                    flag = 0;
                }
            }
            /*@ loop invariant
                //第一种情况：判断是否在同一列上时，已经把flag置为0。
                2<=i<=QUEENS && 1<=k<=i && (\forall integer j:[1..i].0 < a[j] <= QUEENS) && not_finish==1 
                && NotSameDiagonal(a,0,i-1) && NotSameColumn(a,0,i-1) && INT_MAX >= count >= 0 && flag==0 
                ||
                //第二种情况：判断是否在同一列上时，flag仍为1，在判断是否在同一对角线上有可能置为0。
                1<=i<=QUEENS && 1<=k<=i && (\forall integer j:[1..i].0 < a[j] <= QUEENS) && not_finish==1 
                && NotSameDiagonal(a,0,i-1) && NotSameColumn(a,0,i) && INT_MAX >= count >= 0 
                && ( flag==1 && (\forall integer j:[1..k-1].a[i] != a[j]-(j-i) && a[i] != a[j]+(j-i)) 
                  || flag==0 && i > 1 && (a[i] == a[k-1]-(k-1-i) || a[i] == a[k-1]+(k-1-i))
                   )
                ;
            */
            for (k = 1; flag == 1 && k < i; k = k + 1)
            { // 判断是否有多个皇后在同一对角线上。
                if (a[i] == a[k] - (k - i) || a[i] == a[k] + (k - i))
                { // 注意：i==1时，k<i条件不成立，因此flag=0时，必有i>1。
                    flag = 0;
                }
            }
            if (flag == 0)
            { //若存在矛盾，则需要重新设置第i个元素
                if (a[i] == a[i - 1])
                {              //若a[i]的值已经经过一圈追上a[i-1]的值。
                    i = i - 1; //退回一步，重新试探处理前一个元素
                    if (i > 1 && a[i] == QUEENS)
                    {
                        a[i] = 1; //当a[i]为Queens时将a[i]的值置1
                    }
                    else
                    {
                        if (a[i] == QUEENS)
                            not_finish = 0; //当第一行的值达到Queens时结束
                        else
                            a[i] = a[i] + 1; //将a[il的值取下一个值
                    }
                }
                else if (a[i] == QUEENS)
                {
                    a[i] = 1;
                }
                else
                {
                    a[i] = a[i] + 1; //将a[i]的值取下一个值
                }
            }
            else
            { //若不存在矛盾，则试探下一个皇后
                i = i + 1;
                if (i <= QUEENS)
                {
                    if (a[i - 1] == QUEENS)
                        a[i] = 1; // 若前一个元素的值为Queens则a[i]=l
                    else
                        a[i] = a[i - 1] + 1; // 否则元素的值为前一个元素的下一个值
                }
            }
        }
        if (not_finish == 1)
        {
            if (count == INT_MAX)
            { // 解的个数已超出最大整型数的表示范围，退出。
                //@ ghost arithmetic_overflow = true;
                exit(1);
            }
            count = count + 1;
            printf((count - 1) % 3 ? "\t[%2d]:" : "\n[%2d]:", count);
            /*@ loop invariant 
                i==QUEENS+1 && (\forall integer j:[1..QUEENS].0 < a[j] <= QUEENS) && not_finish==1 
                && NotSameColumn(a,0, QUEENS-1) && NotSameDiagonal(a,0, QUEENS-1) 
                && (\forall integer i:[1..QUEENS-1].a[i]!=a[QUEENS]) 
                && (\forall integer i:[1..QUEENS-1].a[QUEENS]!=a[i]-(i-QUEENS) && a[QUEENS]!=a[i]+(i-QUEENS)) 
                && INT_MAX >= count >= 0 && 1<=k<=QUEENS+1
                ;
            */
            for (k = 1; k <= QUEENS; k = k + 1) //输出结果
                printf(" %d", a[k]);
            if (a[QUEENS - 1] < QUEENS)
                a[QUEENS - 1] = a[QUEENS - 1] + 1; // 修改倒数第二位的值
            else
                a[QUEENS - 1] = 1; // 否则从第一列开始
            i = QUEENS - 1;        //开始寻找下一个满足条件的解
        }
    }
    printf("\n");
}